#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List

import logging, itertools

from PySide2.QtWidgets import QWidget

from src.mg_dialog_utils import MgDialogWithRepoList
from src.gui.ui_git_push_tag import Ui_GitPushTag
from src.mg_exec_window import MgExecWindow
from src.mg_repo_info import MgRepoInfo
from src.mg_ensure_info_available import MgEnsureInfoAvailable, RepoInfoFlags

logger = logging.getLogger('mg_dialog_git_push_tag')
dbg = logger.debug

MSG_PUSH_ALL_TAGS = 'Push all tags'

class MgDialogGitPushTag(MgDialogWithRepoList):
    ui: Ui_GitPushTag

    def __init__(self, parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
        super().__init__(parent, Ui_GitPushTag, selectedRepos, allRepos)
        self.sigRepoListAdjusted.connect(self.updateTagContent)
        self.updateTagContent()


    def updateTagContent(self) -> None:
        self.ensureInfoAvailable = MgEnsureInfoAvailable(self, self.getTargetedRepoList(), showProgressDialog=True)
        self.ensureInfoAvailable.ensureInfoAvailable(RepoInfoFlags.ALL_TAGS, blocking=True)
        self.ui.comboBoxTagName.clear()
        self.ui.comboBoxTagName.addItem(MSG_PUSH_ALL_TAGS)
        self.ui.comboBoxTagName.insertSeparator(self.ui.comboBoxTagName.count())
        allTags = list(set(itertools.chain.from_iterable(repo.all_tags for repo in self.getTargetedRepoList())))
        allTags.sort()
        self.ui.comboBoxTagName.addItems(allTags)
        self.ui.comboBoxTagName.setCurrentIndex(0)


    def isPushAllTags(self) -> bool:
        return self.ui.comboBoxTagName.currentIndex() == 0


    def getTagName(self) -> str:
        return self.ui.comboBoxTagName.currentText()


def runDialogGitPushTag(parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
    dbg('runDialogGitPushTag')

    dlg = MgDialogGitPushTag(parent, selectedRepos, allRepos)
    result = dlg.exec()
    if not result:
        # command execution canceled
        return

    if dlg.isPushAllTags():
        desc = 'Pushing all tags'
        cmdPush = ['push', 'origin', '--tags', '--verbose']
    else:
        tagName = dlg.getTagName()
        desc = 'Pushing tag ' + tagName
        cmdPush = ['push', 'origin', 'tag', tagName  ]

    # show window for executing git
    gitExecWindow = MgExecWindow(parent)
    gitExecWindow.execOneGitCommand(desc, cmdPush, dlg.getTargetedRepoList())





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import TYPE_CHECKING, List, Tuple
import tempfile, logging

from PySide2.QtWidgets import QMessageBox, QWidget
from PySide2.QtCore import QRegExp
from PySide2.QtGui import QRegExpValidator

from src.gui.ui_git_tag import Ui_GitAddTag
from src.mg_dialog_utils import reBranchTagValues, MgDialogWithRepoList
from src.mg_exec_window import MgExecWindow
from src.mg_repo_info import MgRepoInfo
from src.mg_exec_task_item import MgExecTaskGit, MgExecTaskGroup
from src import mg_config as mgc

logger = logging.getLogger('mg_dialog_tag')
dbg = logger.debug


class MgDialogGitAddTag(MgDialogWithRepoList):
    ui: Ui_GitAddTag

    def __init__(self, parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
        super().__init__(parent,
                         Ui_GitAddTag,
                         selectedRepos,
                         allRepos)

        # default branch name is empty
        self.ui.comboBoxTagName.insertItem(0, "")
        self.ui.comboBoxTagName.setCurrentIndex(0)
        self.ui.comboBoxTagName.setFocus()
        validator = QRegExpValidator(QRegExp(reBranchTagValues))
        self.ui.comboBoxTagName.setValidator(validator)

        # Fill up list of recent tags
        if len(mgc.get_config_instance().lruAsList(mgc.CONFIG_TAG_HISTORY)):
            self.ui.comboBoxTagName.insertItems(0, mgc.get_config_instance().lruAsList(mgc.CONFIG_TAG_HISTORY))


    def accept(self) -> None:
        tagName = self.ui.comboBoxTagName.currentText()
        if len(tagName) == 0:
            QMessageBox.warning(self, "No tag name specified",
                                "You did not specify any tag name!")
            return

        super().accept()



def runDialogGitTag(parent: QWidget, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
    '''Run git Tag on all listed repos'''
    dbg('runDialogGitTag()')

    # use a dialog to define the tag to set
    dialog = MgDialogGitAddTag(parent, selectedRepos, allRepos)

    result = dialog.exec()
    if not result:
        # tag canceled
        return

    # archive tag name
    tagName = dialog.ui.comboBoxTagName.currentText()
    config = mgc.get_config_instance()
    config.lruSetRecent(mgc.CONFIG_TAG_HISTORY, tagName)
    config.save()

    # check if annotated
    annotation = str(dialog.ui.textAnnotated.toPlainText()).encode('utf8', errors='ignore')

    # contains a list of: (ignore_failure_flag, [git commands to run])
    tag_operation: List[Tuple[bool, List[str]]] = []

    desc = ''
    if annotation:
        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(annotation)
            desc = 'Adding annotated tag ' + tagName
            tag_operation.append( (False, ['tag', tagName, '-F', tmp_file.name] ) )
    else:
        desc = 'Adding lightweight tag ' + tagName
        tag_operation.append( (False, ['tag', tagName] ) )

    if dialog.ui.checkBoxPushToRemote.isChecked():
        desc += ' and pushing it'
        tag_operation.append( (False, ['push', 'origin', tagName]))

    if dialog.ui.checkBoxSwitchToTag.isChecked():
        desc += ' and checkouting it'
        tag_operation.append((True, ['checkout', f'{tagName}~1', '--' ]))
        tag_operation.append((False, ['checkout', tagName, '--']))

    taskGroup = []
    for repo in dialog.getTargetedRepoList():
        taskGroup.append( MgExecTaskGroup(repo.name, repo,
                                          [MgExecTaskGit('', repo, cmdLine, ignore_failure)
                                           for (ignore_failure, cmdLine) in tag_operation]))

    gitExecWindow = MgExecWindow(parent)
    gitExecWindow.execTaskGroups(desc, taskGroup)




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_about.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

# import multigit_resources_rc

class Ui_dialogAbout(object):
    def setupUi(self, dialogAbout: QDialog) -> None:
        if not dialogAbout.objectName():
            dialogAbout.setObjectName(u"dialogAbout")
        dialogAbout.resize(895, 571)
        palette = QPalette()
        brush = QBrush(QColor(255, 255, 255, 255))
        brush.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Active, QPalette.Base, brush)
        palette.setBrush(QPalette.Active, QPalette.Window, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Base, brush)
        palette.setBrush(QPalette.Inactive, QPalette.Window, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Base, brush)
        palette.setBrush(QPalette.Disabled, QPalette.Window, brush)
        dialogAbout.setPalette(palette)
        self.verticalLayout_2 = QVBoxLayout(dialogAbout)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label = QLabel(dialogAbout)
        self.label.setObjectName(u"label")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setMaximumSize(QSize(100, 100))
        self.label.setPixmap(QPixmap(u":/img/multigit-logo-256.png"))
        self.label.setScaledContents(True)
        self.label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)

        self.verticalLayout.addWidget(self.label)


        self.horizontalLayout.addLayout(self.verticalLayout)

        self.textBrowserTitle = QTextBrowser(dialogAbout)
        self.textBrowserTitle.setObjectName(u"textBrowserTitle")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.textBrowserTitle.sizePolicy().hasHeightForWidth())
        self.textBrowserTitle.setSizePolicy(sizePolicy1)
        self.textBrowserTitle.setMinimumSize(QSize(0, 120))
        self.textBrowserTitle.setFocusPolicy(Qt.NoFocus)
        self.textBrowserTitle.setAcceptDrops(False)
        self.textBrowserTitle.setAutoFillBackground(False)
        self.textBrowserTitle.setFrameShape(QFrame.NoFrame)
        self.textBrowserTitle.setFrameShadow(QFrame.Plain)
        self.textBrowserTitle.setLineWidth(0)
        self.textBrowserTitle.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.textBrowserTitle.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.textBrowserTitle.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)

        self.horizontalLayout.addWidget(self.textBrowserTitle)

        self.label_2 = QLabel(dialogAbout)
        self.label_2.setObjectName(u"label_2")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy2)
        self.label_2.setPixmap(QPixmap(u":/img/idemia-logo.png"))

        self.horizontalLayout.addWidget(self.label_2)


        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.textBrowserContent = QTextBrowser(dialogAbout)
        self.textBrowserContent.setObjectName(u"textBrowserContent")
        sizePolicy3 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.textBrowserContent.sizePolicy().hasHeightForWidth())
        self.textBrowserContent.setSizePolicy(sizePolicy3)
        self.textBrowserContent.setFocusPolicy(Qt.NoFocus)
        self.textBrowserContent.setFrameShape(QFrame.NoFrame)
        self.textBrowserContent.setOpenExternalLinks(True)
        self.textBrowserContent.setOpenLinks(True)

        self.verticalLayout_2.addWidget(self.textBrowserContent)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.pushButtonFullLicense = QPushButton(dialogAbout)
        self.pushButtonFullLicense.setObjectName(u"pushButtonFullLicense")
        self.pushButtonFullLicense.setMaximumSize(QSize(16777215, 16777214))
        font = QFont()
        font.setPointSize(10)
        self.pushButtonFullLicense.setFont(font)

        self.horizontalLayout_2.addWidget(self.pushButtonFullLicense)

        self.widget = QWidget(dialogAbout)
        self.widget.setObjectName(u"widget")

        self.horizontalLayout_2.addWidget(self.widget)

        self.pushButtonOk = QPushButton(dialogAbout)
        self.pushButtonOk.setObjectName(u"pushButtonOk")
        self.pushButtonOk.setFont(font)

        self.horizontalLayout_2.addWidget(self.pushButtonOk)


        self.verticalLayout_2.addLayout(self.horizontalLayout_2)

        QWidget.setTabOrder(self.textBrowserTitle, self.textBrowserContent)
        QWidget.setTabOrder(self.textBrowserContent, self.pushButtonFullLicense)
        QWidget.setTabOrder(self.pushButtonFullLicense, self.pushButtonOk)

        self.retranslateUi(dialogAbout)

        self.pushButtonOk.setDefault(True)


        QMetaObject.connectSlotsByName(dialogAbout)
    # setupUi

    def retranslateUi(self, dialogAbout: QDialog) -> None:
        dialogAbout.setWindowTitle(QCoreApplication.translate("dialogAbout", u"About Multigit", None))
        self.label.setText("")
        self.textBrowserTitle.setHtml(QCoreApplication.translate("dialogAbout", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:'MS Shell Dlg 2'; font-size:7.8pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:6.6pt;\"><br /></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:18pt; font-weight:600;\">Multigit by IDEMIA</span></p>\n"
"<p align=\"center\" style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:12pt; font-style:italic;\"><br /></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-"
                        "bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:12pt; font-style:italic;\">Multigit helps managing multiple Git repositories with one interface</span></p></body></html>", None))
        self.label_2.setText("")
        self.textBrowserContent.setHtml(QCoreApplication.translate("dialogAbout", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:'MS Shell Dlg 2'; font-size:7.8pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:6.6pt;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Multigit OpenSource version [version]</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Copyright IDEMIA 2019-2024</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right"
                        ":0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Multigit is developed at </span><a href=\"https://www.idemia.com/\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">IDEMIA</span></a><span style=\" font-size:10pt;\"> by Philippe Fremy.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Multigit is licensed under </span><a href=\"http://www.apache.org/licenses/LICENSE-2.0\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">Apache License 2.0</span></a></p>\n"
"<p style=\"-qt-paragraph-type:em"
                        "pty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Multigit includes the software components:</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">- </span><a href=\"https://www.python.org\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">Python 3.8</span></a><span style=\" font-size:10pt;\">  (from the Python Software Foundation)</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">- </span><a href=\"https://www.qt.io/\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">Qt for Python 5"
                        ".15 </span></a><span style=\" font-size:10pt;\">(by the Qt Group)</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">- </span><a href=\"https://pypi.org/project/concurrent-log-handler/\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">Concurrent-log-handler</span></a><span style=\" font-size:10pt;\"> (by Preston Landers)</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">- </span><a href=\"https://pyinstaller.org/en/stable/\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">PyInstaller</span></a><span style=\" font-size:10pt;\"> (by David Cortesi)</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-size:10pt;\"><br"
                        " /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Most icons are provided by freely </span><a href=\"https://icons8.com\"><span style=\" font-size:10pt; text-decoration: underline; color:#0000ff;\">Icons8</span></a></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:10pt;\">Icons from the softwares Git, TortoiseGit, SourceTree, SublimeMerge are used to represent activation of the related software. Icons remain the properties of their copyright owner. IDEMIA logo is copyrighted to IDEMIA and can not be used without IDEMIA's approval (see licensing for details).</span></p></body></html>", None))
        self.pushButtonFullLicense.setText(QCoreApplication.translate("dialogAbout", u"Full Licensing Information", None))
        self.pushButtonOk.setText(QCoreApplication.translate("dialogAbout", u"Ok", None))
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import Dict, Tuple, Optional, List, Callable, Any, Sequence
import logging, re, csv, pathlib, os
from glob import glob, escape

from PySide2.QtCore import Signal, QObject, QCoreApplication
from PySide2.QtWidgets import QMessageBox

from src.mg_const import MSG_NO_COMMIT, MSG_REMOTE_TOPUSH_TOPULL, MSG_REMOTE_SYNCHRO_OK, MSG_REMOTE_TOPULL, \
    MSG_REMOTE_TOPUSH, MSG_REMOTE_BRANCH_GONE, MSG_LOCAL_BRANCH, SHORT_SHA1_NB_DIGITS, MSG_EMPTY_REPO
from src.mg_tools import RunProcess, ExecGit
from src.mg_utils import anonymise_git_url

logger = logging.getLogger('mg_repo_info')
dbg = logger.debug
warn = logger.warning

def dbg_var(var_name: str, var_value: str) -> None:
    logger.debug('{}="{}"'.format(var_name, var_value))

RECYCLE_BIN=pathlib.Path(r'\$Recycle.Bin')

def is_git_repo(repo: pathlib.Path) -> bool:
    '''Return True if directory is actually a git repo.
    Checks for the presence of .git directory and other mandatory files.
    '''
    git_dir = repo/'.git'
    try:
        rec_bin = pathlib.Path(git_dir.drive)/RECYCLE_BIN
        _ = git_dir.relative_to(rec_bin)
        # No exception raised, we are in the recycle bin
        # we don't want to return git repo from the recycle bin
        return False
    except ValueError:
        # good, git_dir is not in the Recycle Bin...
        pass

    for subd in [ 'hooks', 'info', 'objects', 'refs' ]:
        subd_path = git_dir/subd
        if not subd_path.exists() or not subd_path.is_dir():
            return False
    return True

re_ahead_behind = re.compile('(ahead (?P<ahead2>\\d+), behind (?P<behind2>\\d+))|(ahead (?P<ahead>\\d+))|(behind (?P<behind>\\d+))')
def match_ahead_behind(s: str) -> Tuple[int, int]:
    '''Match a string and return the number for ahead and the number for behind
    Possible matches: "ahead 33", "behind 22", "ahead 33, behind 22"
    '''
    mo = re_ahead_behind.match(s)
    if mo is None:
        warn('No re match for match_ahead_behind("%s")' % s)
        return 0, 0
    nb_ahead, nb_behind = 0, 0
    if mo.group('ahead'):
        nb_ahead = int(mo.group('ahead'))
    elif mo.group('ahead2'):
        nb_ahead = int(mo.group('ahead2'))
    if mo.group('behind'):
        nb_behind = int(mo.group('behind'))
    elif mo.group('behind2'):
        nb_behind = int(mo.group('behind2'))
    return nb_ahead, nb_behind

hexdigits = '0123456789abcdefABCDEF'
def is_not_sha1(ref: str) -> bool:
    '''Return True is this can not be a commit SHA1 and False if this may be a SHA1 or a tag with only numbers+A-F letters'''
    if ref.isdigit():
        # only digits, no way to guess between commit and tags
        return False

    if not ref.isalnum():
        # if contains '.' or '_', this is a tag name
        return True

    non_hex_chars = [c for c in ref if not c in hexdigits]
    if len(non_hex_chars):
        return True

    # we don't know
    return False


class MultiRepo:
    '''Public fields:
    - base_path: base path for finding all contained git repositories
    - repo_dict: dictionnary mapping repo names to MgRepoInfo
    - repo_names: sorted list of repository names.
    '''
    # base_dir an base_path represent the same information in two format: the base directory
    base_path: pathlib.Path
    base_dir: str

    # to iterate over all repo names in sorted order
    repo_names: List[str]

    # to itereate  over all repos objects in sorted order
    repo_list: List['MgRepoInfo']

    # to find a repo given its name
    repo_dict: Dict[str, 'MgRepoInfo']

    def __init__(self, base_dir: str = ''):
        self.base_path = pathlib.Path(base_dir)
        dbg('MultiRepo with dir: %s, %d' % (self.base_path, self.base_path.is_dir()))
        self.base_dir = base_dir
        self.repo_dict = {}
        self.repo_list = []
        self.repo_names = []

    def isEmpty(self) -> bool:
        '''Return whether this instance has been filled or is just empty'''
        return self.base_dir == ''

    def __repr__(self) -> str:
        s = 'MultiRepo< %s >' % self.base_dir
        s += '\n\t' + '\n\t'.join( '%s: %s' % (k, v) for k,v in self.repo_dict.items())
        return s

    def __getitem__(self, item: str) -> 'MgRepoInfo':
        '''Return the MgRepoInfo object from a given repo name'''
        return self.repo_dict[item]

    def __len__(self) -> int:
        '''Return the number of repos'''
        return len(self.repo_names)

    def find_git_repos(self) -> List[str]:
        '''Return a list of git repos contained in base_dir, including base_dir itself, sorted by name'''
        self.repo_dict = {}
        self.repo_list = []
        self.repo_names = []
        self.base_path = pathlib.Path(self.base_dir)
        # Use glob.glob(recursive=True) to follow symlinks whereas Path.glob() doesn't.
        # RND_P_5RNDIT_05-85: ensure base path does not contain magic chars like []*?
        for d in glob(escape(str(self.base_path)) + '/**/.git', recursive=True):
            repo = pathlib.Path(d).parent
            if is_git_repo(repo):
                repo_name = str(repo.relative_to(self.base_path))
                self.repo_names.append(repo_name)

                repo_path = str(repo.resolve())
                repo_info = MgRepoInfo(repo_name, repo_path, repo_name)
                self.repo_dict[repo_name] = repo_info
                self.repo_list.append(repo_info)
                repo_info.repo_deleted.connect(self.slotRepoDeleted)

        self.repo_names.sort(key=lambda name: name.lower())
        self.repo_list.sort(key=lambda repo: repo.name.lower())
        dbg('Repo list:\n%s' % self.repo_names)
        return self.repo_names


    def slotRepoDeleted(self, repo_name: str) -> None:
        '''Called when one of the repo is deleted'''
        dbg('MultiRepo.slotRepoDeleted(%s)' % repo_name)
        if not repo_name in self.repo_dict:
            dbg('repo already deleted!')
            return

        repo = self.repo_dict[repo_name]
        del self.repo_names[self.repo_names.index(repo_name)]
        del self.repo_list[self.repo_list.index(repo)]
        del self.repo_dict[repo_name]


    def find_git_repos_added_removed(self) -> Tuple[List['MgRepoInfo'], List['MgRepoInfo']]:
        '''Return a tuple of:
        - new repos created since the initial scan
        - repos removed since the initial scan

        The content of MultiRepo is adjusted accordingly
        '''
        new_repo_names = []
        self.base_path = pathlib.Path(self.base_dir)
        # Use glob.glob(recursive=True) to follow symlinks whereas Path.glob() doesn't.
        # RND_P_5RNDIT_05-85: ensure base path does not contain magic chars like []*?
        for d in glob(escape(str(self.base_path)) + '/**/.git', recursive=True):
            repo = pathlib.Path(d).parent
            if is_git_repo(repo):
                repo_name = str(repo.relative_to(self.base_path))
                new_repo_names.append(repo_name)

        added_repo_names = set(new_repo_names) - set(self.repo_names)
        rm_repo_names = set(self.repo_names) - set(new_repo_names)
        added_repo = []
        for repo_name in added_repo_names:
            repo_path = str((self.base_path / repo_name).resolve())
            repo_info = MgRepoInfo(repo_name, repo_path, repo_name)
            added_repo.append(repo_info)

        rm_repo = [ self.repo_dict[repo_name] for repo_name in rm_repo_names ]
        return added_repo, rm_repo


    def adjust_git_repos(self, added_repos: List['MgRepoInfo'], rm_repo: List['MgRepoInfo']) -> None:
        '''Adjust the internal list of repos according to the repos added and removed.'''
        for repo in rm_repo:
            del self.repo_names[self.repo_names.index(repo.name)]
            del self.repo_dict[repo.name]
            del self.repo_list[ self.repo_list.index(repo) ]

        for repo in added_repos:
            self.repo_dict[repo.name] = repo
            self.repo_list.append(repo)
            self.repo_names.append(repo.name)
            # don't forget that we want to know when the repo is deleted!
            repo.repo_deleted.connect(self.slotRepoDeleted)

        self.repo_names.sort(key=lambda name: name.lower())
        self.repo_list.sort(key=lambda repo: repo.name.lower())
        dbg('New repo list:\n%s' % self.repo_names)


    def exportCsv(self, fname: str, fieldsRequested: Optional[Dict[str, bool]] = None) -> None:
        '''Export the view on the repo to the CSV'''
        with open(fname, 'w', newline='', errors='replace') as csvf:
            writer = csv.writer(csvf, delimiter=';')
            writer.writerow( ['Path', 'HEAD', 'current branch', 'current tag', 'commit sha1', 'url', 'date'] )

            for repo in sorted(self.repo_list, key=lambda repo: repo.relpath):
                QCoreApplication.processEvents()  # give a chance to asynchronous jobs to complete
                # fetching the information but normally, it was fetched in advance
                repo.ensure_head_and_url_and_commit_date(blocking=True)

                repo_url = anonymise_git_url(repo.url or '')

                if fieldsRequested is None:
                    writer.writerow( [repo.name, repo.head, repo.branch, repo.tag, repo.commit_sha1, repo.commit_date, repo_url] )
                else:
                    content = []
                    content.append(repo.name if fieldsRequested['path'] else '')
                    content.append(repo.head if fieldsRequested['head'] else '')
                    content.append(repo.branch if fieldsRequested['branch'] else '')
                    content.append(repo.tag if fieldsRequested['tag'] else '')
                    content.append((repo.commit_sha1 or '') if fieldsRequested['commit_sha1'] else '')
                    content.append((repo_url or '') if fieldsRequested['url'] else '')
                    content.append((repo.commit_date or '') if fieldsRequested['commit_date'] else '')
                    writer.writerow( content )


class MgRepoInfo(QObject):
    '''Class managing the access to a repository.

    When the class is created, it does not fetch any information.

    Use:
    - fill_repo_info() to fill the base information about a repository
    - fill_repo_url() to also fill the url part
    - fill_last_commit() to fill the last commit part.
    '''

    # Internal conventions:
    # last_commit: empty when not filled, not empty when filled => always a string

    name: str                   # name of the repo == relative path to the base directory
    relpath: str                # the relative path of this repo to the base dir, usually same as name
    fullpath: str               # the full path to the repository, used for running git commands
    url: Optional[str]          # None when not filled yet, string (possibly empty) when filled
    status: str                 # emtpy string when not filled, string when filled. Description of the modified files
    tag: str                    # current tag we are on, if any. Empty string when not pointing on a tag
    branch: str                 # current branch we are on, if any. Emtpy string when not on a branch, "<empty repo>" is also a valid value
    head: str                   # description of what head points: "branch XXX", "tag XXX", "commit XXX", "<empty repo>"
    last_commit: str            # message of the last commit
    commit_date: Optional[str]  # date of the last commit
    commit_sha1: Optional[str]  # sha1 of the last commit, None if not set
    diff: Optional[str]         # diff of the last commit if any
    diff_summary: Optional[str] # list of modified files in the diff
    remote_branch: str          # remote-branch of any, or empty string
    remote_synchro: str         # informative message about sync with remote branch:
                                    # "xx to push, yy to pull"
                                    # "remote branch gone"
                                    # "NA for local branch"
                                    # "up-to-date"
    branches_local: List[str]   # list of all local branches on this repo
    branches_remote: List[str]  # list of all remote branches on this repo
    branches_filled: bool       # True when the attributes branches_* have been filled
    tags: Optional[str]         # all the tags pointing at this commit. None when not filled, comma separated list of tags
    all_tags: List[str]   # all tags existing on this repo
    all_tags_filled: bool           # set when all tags of the repo
    files_sha1: List[Tuple[str, str]]  # List of (files, sha1)
    files_sha1_filled: bool           # set when files sha1 of the repo is filled
    is_deleted: bool                # set to True when emitting the signal repo_deleted

    tag_or_commit: Optional[str]    # internal field, used in log -1
    force_blocking_git: bool        # internal field, used in ensure_all_filled()

    # signals to emit
    repo_update_in_progress = Signal(str)
    repo_info_available = Signal(str)
    repo_deleted = Signal(str)


    def __repr__(self) -> str:
        s = 'MgRepoInfo< %s >' % self.name
        return s


    def __init__(self, name: str, fullpath: str, relpath: str = '') -> None:
        super().__init__()

        self.name = name
        self.relpath = relpath
        self.force_blocking_git = False

        self.cb_repo_info_available: Optional[Callable[[str], Any]] = None

        # full path to our repo
        self.fullpath = fullpath

        self._clear_all()

    def _clear_all(self, clearUrl: bool = True) -> None:
        dbg(f'clear_all(clearUrl={clearUrl}) - {self.name}')
        self._clear_basic_info()
        if clearUrl:
            self.url = None
        self.last_commit = ''
        self.diff = None
        self.diff_summary = None
        self.commit_sha1 = None
        self.tag_or_commit = None
        self.commit_date = None
        self.branches_filled = False
        self.branches_local = []
        self.branches_remote = []
        self.is_deleted = False


    def abortBecauseRepoDeleted(self) -> bool:
        '''Called to check if repo has been deleted.
        Emits the appropriate signal and return True if repo was deleted.
        This is to be called each time git returns an error code'''
        if not os.path.exists(self.fullpath)  \
            or not os.path.exists(os.path.join(self.fullpath, '.git')):
            self.is_deleted = True
            self.repo_deleted.emit(self.relpath)
            return True

        return False


    def _clear_basic_info(self) -> None:
        '''Clear the basic information on a repo, displayed in the main window: head, branch, tag, status, remote_branch, tags'''
        dbg('clear_basic_info() - %s' % self.name)
        self.branch = ''
        self.tag = ''
        self.head = ''
        self.remote_branch = ''
        self.status = ''
        self.remote_synchro = ''
        self.tags = None
        self.all_tags = []
        self.all_tags_filled = False
        self.files_sha1 = []
        self.files_sha1_filled = False


    def deepRefresh(self) -> 'MgRepoInfo':
        '''Reread all relevant information from git repositories, erasing remote url '''
        dbg('deepRefresh() - %s' % self.name)
        return self._refresh(True)


    def refresh(self) -> 'MgRepoInfo':
        '''Reread all relevant information from git repositories, keeping remote url'''
        dbg('refresh() - %s' % self.name)
        return self._refresh(False)


    def _refresh(self, clearUrl: bool) -> 'MgRepoInfo':
        '''Reread all relevant information from git repositories, and keep or clear url depending on parameter'''
        dbg(f'_refresh(clearUrl={clearUrl}) - %s' % self.name)
        self._clear_all(clearUrl)
        self.fill_repo_info()
        return self


    def nice_status(self) -> str:
        '''Nicer status line, longer to display'''
        s = self.status
        return s


    def repo_info_is_available(self) -> None:
        '''Called when basic information has been filled (result of calling fill_repo_info()).

        Actions:
        * calls the callback self.cb_repo_info_available if set, and reset it to None
        * emit the signal repo_info_available
        '''
        if self.cb_repo_info_available is not None:
            # set self.cb_repo_info_available to None first, to avoid recursive calls
            cb_repo_info_available = self.cb_repo_info_available
            self.cb_repo_info_available = None
            cb_repo_info_available(self.name)

        self.repo_info_available.emit(self.name)


    def ensure_all_filled(self) -> 'MgRepoInfo':
        '''Ensure that all information of the repo, including URL is filled.

        To be called when displaying properties for example
        '''
        dbg('ensure_all_filled() - %s' % self.name)

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = True

            if self.status == '':
                # not filled yet
                self.fill_repo_info()

            self.ensure_url()
            self.ensure_last_commit()
            self.ensure_tags()
            self.ensure_diff_summary()
            self.ensure_diff()
            self.ensure_commit_date()
        finally:
            self.force_blocking_git = old_force_blocking_git
        return self


    def ensure_head_and_url_and_commit_date(self, blocking: bool = False) -> None:
        '''Ensure that the fields 'head' and 'url' and 'commit_date' are filled.
        If blocking is True, the call is blocking. Else, an asynchronous call is made'''
        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking

            if self.head == '':
                self.fill_repo_info()
            self.ensure_url()
            self.ensure_commit_date()
        finally:
            self.force_blocking_git = old_force_blocking_git


    def ensure_head_and_url(self, blocking: bool = False, cb_info_and_url_avail: Optional[Callable[[str], Any]] = None) -> None:
        '''Ensure that the fields 'head' and 'url' are filled.
        If blocking is True, the call is blocking. Else, an asynchronous call is made'''
        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking

            def local_cb_url(_url: str) -> None:
                if cb_info_and_url_avail is not None:
                    cb_info_and_url_avail(self.name)

            def local_cb_fill_repo_info(_repo_name: str) -> None:
                # called when repository information is available
                self.ensure_url(local_cb_url)

            # basic information is missing
            if self.head == '':
                self.fill_repo_info(local_cb_fill_repo_info)
                return

            self.ensure_url(local_cb_url)

        finally:
            self.force_blocking_git = old_force_blocking_git


    def ensure_url(self, cb_url: Optional[Callable[[str], Any]] = None) -> None:
        '''Ensure that url field is correctly filled. If provided callback is called synchronously or asynchronously
        with the URL'''
        if self.url is not None:
            if cb_url:
                cb_url(self.url)
            return

        self.fill_repo_url(cb_url)


    def ensure_last_commit(self, cb_last_commit: Optional[Callable[[str, str], Any]] = None) -> None:
        '''Ensure that the last_commit field gets filled

        Calls cb_last_commit() when the last_commit information is available. This
        is either immediate or pending the git command execution.

        The call to cb_last_commit is done with: cb_last_commit(<repo_name>, <last_commit_content>)
        '''
        dbg('ensure_last_commit() - %s' % self.name)
        if self.last_commit:
            if cb_last_commit:
                cb_last_commit(self.name, self.last_commit)
            return

        self.fill_last_commit(cb_last_commit)


    def fill_last_commit(self, cb_last_commit: Optional[Callable[[str, str], Any]] = None) -> None:
        '''Force the filling of the fields:
         - last_commit
         - commit_date (side-effect)
         - commit_sha1 (side-effect)
         '''

        # local function to allow calling the callback
        def local_fill_last_commit_git_log_done(repo_name: str, git_exit_code: int, git_output: str) -> None:
            self.cb_fill_repo_info_log_done(repo_name, git_exit_code, git_output)
            if cb_last_commit:
                cb_last_commit(repo_name, self.last_commit)

        # errors possible for empty repository
        self.git_exec_async_here(['log', '-1'], local_fill_last_commit_git_log_done, allow_errors=True)


    def ensure_diff(self, cb_fill_diff: Optional[Callable[[str, str], Any]] = None) -> None:
        '''Ensure that diff information is filled'''
        dbg('ensure_diff() - %s' % self.name)
        if self.diff is not None:
            if cb_fill_diff:
                cb_fill_diff(self.name, self.diff)
            return

        def local_cb_fill_diff_done(repo_name: str, exit_code: int, diff_out: str) -> None:
            if exit_code != 0:
                if self.abortBecauseRepoDeleted():
                    return
                self.show_error_message_bad_git_exit_code(exit_code, diff_out)
            self.diff = diff_out
            if cb_fill_diff:
                cb_fill_diff(repo_name, self.diff)

        # errors possible when repo is deleted
        self.git_exec_async_here(['diff', '-u', '--patience', '--stat'], local_cb_fill_diff_done, allow_errors=True)


    def ensure_diff_summary(self, cb_fill_diff_summary: Optional[Callable[[str, str], Any]] = None, blocking: bool = False) -> None:
        '''Ensure that diff summary is filled'''
        dbg('ensure_diff_summary() - %s' % self.name)
        if self.diff_summary is not None:
            if cb_fill_diff_summary:
                cb_fill_diff_summary(self.name, self.diff_summary)
            return

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking

            def local_cb_git_diff_stat_done(repo_name: str, exit_code: int, diff_out: str) -> None:
                if exit_code != 0:
                    if self.abortBecauseRepoDeleted():
                        return
                    self.show_error_message_bad_git_exit_code(exit_code, diff_out)
                    return

                self.diff_summary = diff_out
                if cb_fill_diff_summary:
                    cb_fill_diff_summary(repo_name, self.diff_summary)

            # errors possible when repo is deleted
            self.git_exec_async_here(['diff', '--stat'], local_cb_git_diff_stat_done, allow_errors=True)
        finally:
            self.force_blocking_git = old_force_blocking_git


    def ensure_commit_date(self) -> None:
        '''Ensure that field commit_date is filled'''
        if self.commit_date is None:
            self.fill_last_commit()   # fill fill self.last_date as a side-effect


    def ensure_tags(self) -> None:
        '''Ensure that fields are filled: tags'''
        dbg('ensure_tags() - %s' % self.name)
        if self.tags is None:
            self.fill_tags()


    def ensure_sha1(self, cb_commit_sha1: Optional[Callable[[str], Any]] = None, blocking: bool = False) -> None:
        '''Ensure that the sha1 information is available. If blocking is True,
        the function will return only after the sha1 is available.'''

        if self.commit_sha1 is not None:
            # we already have it
            if cb_commit_sha1:
                cb_commit_sha1(self.commit_sha1)
            return

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking

            local_cb_last_commit: Optional[Callable[[str, str], None]]
            if cb_commit_sha1:
                def local_cb_last_commit(_repoName: str, _lastCommit: str) -> None:
                    cb_commit_sha1(self.commit_sha1 or '')
            else:
                local_cb_last_commit = None

            # request sha1 with log -1
            self.fill_last_commit(local_cb_last_commit)

        finally:
            self.force_blocking_git = old_force_blocking_git


    def fill_tags(self) -> None:
        '''Fill the following fields: tags (blocking)'''
        dbg('fill_tags() - %s' % self.name)
        if self.last_commit == MSG_NO_COMMIT:
            # no commits means no tags...
            self.tags = ''
            return
        self.tags = ' '.join(self.git_exec_blocking_here('tag', '--points-at', 'HEAD').strip().split('\n'))


    def fill_repo_url(self, cb_url: Optional[Callable[[str], Any]] = None) -> None:
        '''Fill the URL part of the repo: field url and calls the callback if any when done'''
        dbg('fill_repo_url() - %s' % self.name)

        if cb_url:
            def local_cb_fill_git_remote_done(repo_name: str, git_exit_code: int, remote_out: str) -> None:
                self.cb_fill_git_remote_done(repo_name, git_exit_code, remote_out)
                cb_url(self.url or '')
        else:
            local_cb_fill_git_remote_done = self.cb_fill_git_remote_done

        # errors possible when repo is deleted
        self.git_exec_async_here(['remote', '--verbose'], local_cb_fill_git_remote_done, allow_errors=True)


    def cb_fill_git_remote_done(self, repo_name: str, git_exit_code: int, remote_out: str) -> None:
        self.url = ''
        '''Output looks like:       
        origin  https://toto@titi.org/toto/multigit.git (fetch)
        origin  https://toto@titi.org/toto/multigit.git (push)

        or nothing if no remote has been set
        or more lines if there are multiple remotes
        '''
        if git_exit_code != 0:
            if not self.abortBecauseRepoDeleted():
                self.show_error_message_bad_git_exit_code(git_exit_code, remote_out)
            return

        # We give priority to origin and (fetch)
        remote_out_lines = [ l.split(maxsplit=1) for l in remote_out.split('\n') if len(l) ]
        if len(remote_out_lines) == 0:
            return
        origin_url = [ url for (name, url) in remote_out_lines if 'origin' == name]
        if len(origin_url) == 0:
            # no origin, uncommon!
            # but there are other lines to interpret, try them
            origin_url = [url for (name, url) in remote_out_lines]
        url_method = [ l.rsplit(' ', maxsplit=1) for l in origin_url ]
        fetch_url = [ url for url,method in url_method if method == '(fetch)' ]
        assert len(fetch_url) > 0
        self.url = fetch_url[0]


    def fill_repo_info(self, cb_repo_info_available: Optional[Callable[[str], Any]] = None) -> None:
        '''Fill the following fields:
        - status
        - head
        - branch
        - tag
        - remote_synchro

        May also fill:
        - diff
        - last_commit
        - commit_date
        - commit_sha1
        - tags
        '''
        dbg('fill_repo_info() - %s' % self.name)
        self._clear_basic_info()
        self.cb_repo_info_available = cb_repo_info_available
        self.repo_update_in_progress.emit(self.name)
        # errors possible when repo is deleted
        self.git_exec_async_here(['status', '--porcelain', '--branch'], self.cb_fill_repo_info_status_done,
                                 allow_errors=True)


    re_status_mod_files = re.compile('([MDRCUA ][MDRCUA]|[MDRCUA][MDRCUA ]) (.+)')


    def cb_fill_repo_info_status_done(self, _repo_name: str, git_exit_code: int, status_out: str) -> None:
        '''Called after git status'''
        dbg('fill_repo_info_status_done() - %s, exit_code=%d' % (self.name, git_exit_code))
        if git_exit_code != 0:
            if not self.abortBecauseRepoDeleted():
                self.show_error_message_bad_git_exit_code(git_exit_code, status_out)
            return

        '''git status output:
        "## No commits yet on master
            -> no commit yet
        "## Initial commit on master" 
            -> ???
        "## <branch name>" 
            -> on a branch, no remote
        "## <branch name>...<remote branch name>" 
            -> on a branch, with remote, up-to-date
        "## <branch name>...<remote branch name> [behind N]" 
            -> on a branch, with remote, not uptodate
        "## <branch name>...<remote branch name> [ahead N]" 
            -> on a branch, with remote, not uptodate
        "## HEAD (no branch)" 
            -> detached head

        "?? filename"
            -> untracked files

        " M <fname with or without quotes>"
        " D <fname>"
        " R <fname>"
        " C <fname>"
        " U <fname>"
            -> modified file name
        " A <fname> -> <fname>"
            -> renamed file name
        '''
        status_out_lines = status_out.split('\n')
        mod_files = [ self.re_status_mod_files.match(l).group(2)  # type: ignore # mypy does not understand theat the None value is excluded from the list
                        for l in status_out_lines 
                        if self.re_status_mod_files.match(l) 
        ]
        
        if len(mod_files) == 1:
            self.status = '{} modified file'.format(len(mod_files))
        elif len(mod_files) > 1:
            self.status = '{} modified files'.format(len(mod_files))
        else:
            self.status = 'OK'
            self.diff = ''

        if '## HEAD (no branch)' in status_out:
            # HEAD is detached
            self.git_exec_async_here(['branch'], self.cb_fill_repo_info_branch_done)
            return

        # HEAD is not detached, we are on a branch or on an empty repo

        if '## No commits yet on' in status_out or \
            '## Initial commit on' in status_out:
            # repo is only initialized, no commit, no branch yet:
            #   output: "## No commits yet on master"
            # or repo is cloned from an empty repo:
            #   output: "## No commits yet on master...origin/master [gone]"
            self.branch = MSG_EMPTY_REPO
            self.head = MSG_EMPTY_REPO
            self.tags = ''
            self.commit_sha1 = ''
            self.commit_date = ''
            self.last_commit = MSG_NO_COMMIT
            self.remote_synchro = ''
            self.repo_info_is_available()
            return

        # HEAD is not detached, we are on a branch, we can fetch its name from status
        # we can also fetch the remote name, and the status vs remote branch (ahead, behind)
        status_out_header = [l for l in status_out_lines if l.startswith('##') ][0]
        if '...' in status_out_header:
            # branch + remote-tracking branch
            self.branch, remote_branch_and_remote_status = status_out_header[3:].split('...')
            if '[' in remote_branch_and_remote_status:
                # ahead / behind info is present
                self.remote_branch, remote_status = remote_branch_and_remote_status.split(' [')
                remote_status = remote_status.strip(']')
                # remote_status is either: "ahead 33", "behind 2", "ahead 2, behind 3"
                nb_ahead, nb_behind = match_ahead_behind(remote_status)
                if nb_ahead > 0 and nb_behind > 0:
                    self.remote_synchro = MSG_REMOTE_TOPUSH_TOPULL % (nb_ahead, nb_behind)
                elif nb_ahead > 0:
                    self.remote_synchro = MSG_REMOTE_TOPUSH % nb_ahead
                elif nb_behind > 0:
                    self.remote_synchro = MSG_REMOTE_TOPULL % nb_behind
                elif remote_status == 'gone':
                    # possible cases:
                    # - cloning an empty repo
                    #     output: '## No commits yet on master...origin/master [gone]
                    #     => this is already handled at the beginninig of this function, when looking for
                    #        the string "No commits yet on master". So no need to handle it here
                    # - cloning from a branch, then the remote branch is deleted
                    #     output:  ## master...origin/master [gone]

                    # happens when the tracking branch is gone, or when cloning an empty repo
                    # and then comitting to it
                    self.remote_synchro = MSG_REMOTE_BRANCH_GONE
                else:
                    assert False, "neither ahead nor behind are positive: ahead = %d, behind = %d" % (nb_ahead, nb_behind)
            else:
                self.remote_branch = remote_branch_and_remote_status
                self.remote_synchro = MSG_REMOTE_SYNCHRO_OK

            self.head = 'branch %s' % self.branch
        else:
            self.branch = status_out_header[3:]
            self.head = 'branch %s' % self.branch
            self.remote_synchro = MSG_LOCAL_BRANCH

        self.repo_info_is_available()
        return


    def cb_fill_repo_info_branch_done(self, _repo_name: str, git_exit_code: int, branch_out: str) -> None:
        '''Called after "git branch" to fetch the status of the local and remote branch more precisely'''
        dbg('fill_repo_info_branch_done() - %s' % self.name)
        if git_exit_code != 0:
            # error already notified to user by previous callback
            self.abortBecauseRepoDeleted()
            return

        '''git branch output:
        * (no branch)        
            -> no information provided on where the head is detached
        * (HEAD detached at XXX)
            -> tag-name or commit sha, as used directly by checkout
        * (HEAD detached from XXX)
            -> detached head, then some commits occured
        * master
            -> on branch master
          other_branch
            -> other branch also exists
        '''
        branch_out_lines = branch_out.split('\n')

        # if we need to call "git log -1", we need a local function callback to emit the signal repo_info_available
        def local_fill_repo_info_log_done(repo_name: str, git_exit_code: int, git_log_out: str) -> None:
            self.cb_fill_repo_info_log_done(repo_name, git_exit_code, git_log_out)
            self.repo_info_is_available()

        if '* (HEAD detached at' in branch_out:
            # branch output: * (HEAD detached at XXX) => tag-name or commit, used directly by checkout
            tag_line = [l for l in branch_out_lines if 'HEAD detached' in l]
            self.tag_or_commit = tag_line[0][20:-1]
            if is_not_sha1(self.tag_or_commit):
                # this is a tag
                self.tag = self.tag_or_commit
                self.head = 'tag %s' % self.tag
            else:
                # we can not tell if we are on a commit SHA1 or a tag
                # check using the commit sha1
                self.git_exec_async_here(['log', '-1'], local_fill_repo_info_log_done, allow_errors=True)
                return

            self.repo_info_is_available()
            return

        if '* (HEAD detached from' in branch_out or '(no branch)' in branch_out:
            # branch output: * (HEAD detached from XXX) => checkout + commit has occured
            # branch output: * (no branch) => no information provided on where we are detached
            # use git-log to fetch the commit
            self.git_exec_async_here(['log', '-1'], local_fill_repo_info_log_done, allow_errors=True)
            return

        # we know there is one branch there
        br_line = [l for l in branch_out.split('\n') if '*' in l]
        self.branch = br_line[0][2:]
        self.head = 'branch %s' % self.branch
        self.repo_info_is_available()
        return


    def cb_fill_repo_info_log_done(self, _repo_name: str, git_exit_code: int, git_log_out: str) -> None:
        '''Called after "git log -1" to parse the result'''
        dbg('fill_repo_info_log_done() - %s' % self.name)
        # git returns an error on empty repo, don't bother. Else log -1 should always work
        if git_exit_code != 0:
            if 'does not have any commits yet' in git_log_out:
                self.last_commit = MSG_NO_COMMIT
                self.commit_date = ''
                self.commit_sha1 = ''
            else:
                if self.abortBecauseRepoDeleted():
                    return

                nice_git_cmd = '> "git" "-C" "%s" "log" "-1"' % self.fullpath
                # noinspection PyTypeChecker
                msg = nice_git_cmd + '\n' + git_log_out

                self.show_error_message_bad_git_exit_code(git_exit_code, msg)
            return

        self.last_commit = git_log_out
        log_out_lines = git_log_out.split('\n')

        for l in log_out_lines:
            if l.startswith('Date:'):
                self.commit_date = l[6:].strip()
                break
        else:
            warn('Could not locate the Date line inside the log -1 response')

        self.commit_sha1 = log_out_lines[0].split(' ')[1].strip()
        if self.tag_or_commit:
            # we had the message "detached at XXX" and we try to figure out
            # if XXX is a commit or a tag
            if self.tag_or_commit in self.commit_sha1:
                # this was a commit
                self.head = 'commit %s' % self.tag_or_commit[:SHORT_SHA1_NB_DIGITS]
            else:
                # this was a tag
                self.head = 'tag %s' % self.tag_or_commit
                self.tag = self.tag_or_commit
            self.tag_or_commit = None

        elif self.head == '':
            # head is not set yet, we were on a detached head with no information
            self.head = 'commit %s' % self.commit_sha1[:SHORT_SHA1_NB_DIGITS]
        else:
            # we are just parsing the output of log -1, nothing special to do here
            pass


    def ensure_branches_filled(self, cb_branches_filled_done: Optional[Callable[[str], None]] = None, blocking: bool = False) -> None:
        '''Ensure that fields branch_local and branch_remote are filled.

        If provided, the callback function is called with the name of the repository'''
        dbg('ensure_branches_filled() - %s' % self.name)
        if self.branches_filled:
            if cb_branches_filled_done:
                cb_branches_filled_done(self.name)
            return

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking
            self.fill_branches(cb_branches_filled_done)
        finally:
            self.force_blocking_git = old_force_blocking_git



    def ensure_all_tags_filled(self, cb_all_tags_filled_done: Optional[Callable[[str], None]] = None, blocking: bool = False) -> None:
        '''Ensure that fields branch_local and branch_remote are filled.

        If provided, the callback function is called with the name of the repository'''
        dbg('ensure_all_tags_filled() - %s' % self.name)
        if self.all_tags_filled:
            if cb_all_tags_filled_done:
                cb_all_tags_filled_done(self.name)
            return

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking
            self.fill_all_tags(cb_all_tags_filled_done)

        finally:
            self.force_blocking_git = old_force_blocking_git




    def fill_branches(self, cb_branches_filled_done: Optional[Callable[[str], None]] = None) -> None:
        '''Fills fields branch_local and branch_remote are filled.

        If provided, the callback function is called with the name of the repository'''

        def local_cb_fill_branches_done(repo_name: str, git_exit_code: int, git_output: str) -> None:
            self.cb_fill_branches_done(repo_name, git_exit_code, git_output)
            if cb_branches_filled_done:
                cb_branches_filled_done(repo_name)

        # errors possible when repo is deleted
        self.git_exec_async_here(['branch', '--all'], local_cb_fill_branches_done, allow_errors=True)


    def cb_fill_branches_done(self, _repo_name: str, git_exit_code: int, git_output: str) -> None:
        '''Called when [git branch --all] completes'''

        '''Example of output:
  dev
  feat/GF_feature_finish
* feat/improv-misc
  master
  remotes/origin/HEAD -> origin/fix/toto
  remotes/origin/b2
  remotes/origin/dev
  remotes/origin/feat/Clone_from_Project_Configuration_file
  remotes/origin/feat/GF_feature_finish
  
 
        Other example: 
* (HEAD detached at tag1)
  master
        '''

        if git_exit_code != 0:
            if not self.abortBecauseRepoDeleted():
                self.show_error_message_bad_git_exit_code(git_exit_code, git_output)
            return

        self.branches_local = []
        self.branches_remote = []
        for v in git_output.split('\n'):
            v = v.strip()
            if not len(v):
                continue
            if v.startswith('* ('):
                # this is not a branch, just an informative message
                continue
            if v.startswith('* '):
                v = v[2:]
            if ' -> ' in v:
                v = v.split(' -> ')[0].strip()

            if not v.startswith('remotes/'):
                self.branches_local.append(v)
            elif v != 'remotes/origin/HEAD':
                # we skip the HEAD branch for remotes, it is not useful for git usage
                self.branches_remote.append(v[8:])

        self.branches_filled = True


    def fill_all_tags(self, cb_all_tags_filled_done: Optional[Callable[[str], None]] = None) -> None:
        '''Fill the following fields: all_tags'''
        dbg('fill_all_tags() - %s' % self.name)
        if self.last_commit == MSG_NO_COMMIT:
            # no commits means no tags...
            self.all_tags = []
            if cb_all_tags_filled_done:
                cb_all_tags_filled_done(self.name)
            return

        def local_cb_fill_all_tags_done(repo_name: str, git_exit_code: int, git_output: str) -> None:
            self.cb_fill_all_tags_done(repo_name, git_exit_code, git_output)
            if cb_all_tags_filled_done:
                cb_all_tags_filled_done(repo_name)

        self.git_exec_async_here(['tag', '--list', '--sort', 'creatordate'], local_cb_fill_all_tags_done, allow_errors=True)


    def cb_fill_all_tags_done(self, _repo_name: str, git_exit_code: int, git_output: str) -> None:
        if git_exit_code != 0:
            if not self.abortBecauseRepoDeleted():
                self.show_error_message_bad_git_exit_code(git_exit_code, git_output)
            return

        self.all_tags = []
        for v in git_output.split('\n'):
            v = v.strip()
            if not len(v):
                continue

            self.all_tags.append(v)

        self.all_tags_filled = True


    def has_commits_to_push(self) -> bool:
        '''Check if remote synchro has commits to push.

        This is used to check for some git operations invalid when there are commits to push'''
        return 'to push' in self.remote_synchro


    def ensure_files_sha1_filled(self, cb_fill_files_sha1_done: Optional[Callable[[str], None]] = None, blocking: bool = False) -> None:
        '''Ensure that fields files_sha1 is filled.

        If provided, the callback function is called with the name of the repository'''
        dbg('ensure_files_sha1_filled() - %s' % self.name)
        if self.files_sha1_filled:
            if cb_fill_files_sha1_done:
                cb_fill_files_sha1_done(self.name)
            return

        old_force_blocking_git = self.force_blocking_git
        try:
            # force the calls to blocking to make sure information is available after this call
            # opening properties is still not asynchronous
            self.force_blocking_git = blocking
            self.fill_files_sha1(cb_fill_files_sha1_done)

        finally:
            self.force_blocking_git = old_force_blocking_git


    def fill_files_sha1(self, cb_fill_files_sha1_done: Optional[Callable[[str], None]] = None) -> None:
        '''Fill the URL part of the repo: field url and calls the callback if any when done'''
        dbg('fill_files_sha1() - %s' % self.name)

        def local_cb_fill_files_sha1_done(repo_name: str, git_exit_code: int, remote_out: str) -> None:
            self.cb_fill_files_sha1_done(repo_name, git_exit_code, remote_out)
            if cb_fill_files_sha1_done:
                cb_fill_files_sha1_done(self.name)

        # errors possible when repo is deleted
        self.git_exec_async_here(['ls-files', '-s'], local_cb_fill_files_sha1_done, allow_errors=True)


    def cb_fill_files_sha1_done(self, _repo_name: str, git_exit_code: int, git_output: str) -> None:
        if git_exit_code != 0:
            if not self.abortBecauseRepoDeleted():
                self.show_error_message_bad_git_exit_code(git_exit_code, git_output)
            return

        self.files_sha1 = []
        for line in git_output.split('\n'):
            line = line.strip('\t \n')
            if not len(line):
                continue

            _0, sha1, _2, fname, *_3 = line.replace('\t', ' ').split(' ')
            self.files_sha1.append((fname, sha1))

        self.files_sha1.sort()
        self.files_sha1_filled = True



    ##############################################################################
    #
    #          Generic interactions with git
    #
    ##############################################################################


    def git_exec_blocking_here(self, *args: str) -> str:
        '''Execute a git command in the context of the current repo.
        Blocks until the command is complete and returns the command output.

        May raise subprocess.CalledProcessError() if git does not exit with 0
        '''
        if not ExecGit.checkFound():
            return ''

        git_cmd = ['-C', self.fullpath] + list(args)
        return ExecGit.exec_blocking(git_cmd)


    def git_exec_async_here(self, args: Sequence[str], cb_git_done: Optional[Callable[[str, int, str], Any]],
                            allow_errors: bool = False) -> None:
        '''Execute git on the current repo with a callback for when the command is over.

        Note that if self.force_blocking_git is True, the call will be turned into
        a blocking call instead of an async call. The callback is called in both situations.

        if allow_errors is False, a message box is displayed if git returns an exit code different than 0. If you
        have your own handling of git errors, set allow_errors to True.
        '''
        if not ExecGit.checkFound():
            return

        git_cmd = [ExecGit.get_executable(), '-C', self.fullpath] + list(args)

        cb_process_done = None
        if cb_git_done:
            # adapt the callback by including the repo name
            def cb_process_done(git_exit_code: int, git_output: str) -> None:
                cb_git_done(self.name, git_exit_code, git_output)

        # our pool process does not handle force blocking, launch directly
        RunProcess().exec_async(git_cmd, cb_process_done, force_blocking=self.force_blocking_git, allow_errors=allow_errors)


    def show_error_message_bad_git_exit_code(self, git_exit_code: int, git_output: str) -> None:
        QMessageBox.warning(None, 'Error when running git',
                            f'Git bad exit code {git_exit_code}.\n\nError message:\n{git_output}')





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import os
import platform
import sys, logging, datetime, tempfile, pathlib
import logging.handlers
import traceback as tb_module
from typing import Type, Optional
from types import TracebackType

from concurrent_log_handler import ConcurrentRotatingFileHandler

from PySide2.QtWidgets import QApplication, QMessageBox
from PySide2 import QtGui

if __package__:
    # when running inside a package, src/* is not on the path, only multigit_gx is.
    # we must add it explicitely
    path = os.path.dirname(__file__)
    sys.path.insert(0, path)

import src.multigit_resources_rc        # uses side-effects to make resources available
from src.mg_window import MgMainWindow
from src import mg_const

# by default Qt abort on Python exceptions so we need to provide
# our own hook that does the job
def handle_exception(type_: Type[BaseException], value: BaseException, traceback: Optional[TracebackType]) -> None:
    # note: the argument names have to have the same name as in the stdlib
    # else mypy will complain.
    if issubclass(type_, KeyboardInterrupt) and traceback is not None:
        sys.__excepthook__(type_, value, traceback)
        return

    # yeah ... logging also raises exceptions sometimes...
    # noinspection PyBroadException
    try:
        logging.error("Uncaught exception", exc_info=(type_, value, traceback))
    except Exception:
        pass

    if app:
        exc_msg = ''.join( tb_module.format_exception(type_, value, traceback) )
        # noinspection PyTypeChecker
        QMessageBox.critical(None, 'Fatal error',
'''A fatal error occured:
%s\n
Please report the problem on opening an issue on https://github.com/idemia/multigit/ .
Please include the file log_multigit_debug.log which you can find in the menu About / Show Multigit log files .
''' % exc_msg )

    # do nothing else, it's enough already



# mandatory to avoid Python crashing on exceptions raised inside slots
app: Optional[ QApplication ] = None

def main_gui() -> None:
    global app
    app = QApplication([])
    icon = QtGui.QIcon()
    icon.addPixmap(QtGui.QPixmap(":/img/multigit-logo-256.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
    icon.addPixmap(QtGui.QPixmap(":/img/multigit-logo-16.png"),  QtGui.QIcon.Normal, QtGui.QIcon.Off)
    app.setWindowIcon(icon)
    w = MgMainWindow(sys.argv)
    w.setWindowIcon(icon)
    w.show()

    app.exec_()

def is_writeable(fname: pathlib.Path) -> bool:
    '''Return true if file may be written'''
    try:
        with open(fname, 'a+') as _:
            pass
        return True
    except PermissionError:
        return False


def configure_logpath(debug_activated: bool = False, run_from_tests: bool = False) -> None:
    '''Creates the log path in mg_const according to platform and ability to write to directory'''

    if run_from_tests:
        fname_log_normal = 'log_tests_multigit.log'
        fname_log_debug = 'log_tests_multigit_debug.log'
        fname_log_git_cmd = 'log_tests_git_cmd.log'
    else:
        fname_log_normal = 'log_multigit.log'
        fname_log_debug  = 'log_multigit_debug.log'
        fname_log_git_cmd = 'log_git_cmd.log'

    path_log_normal: Optional[pathlib.Path]
    path_log_debug: Optional[pathlib.Path]
    path_log_git_cmd: Optional[pathlib.Path]

    if sys.platform == 'win32':
        multigit_log_dir = pathlib.Path(os.environ['USERPROFILE']) / 'AppData/Local/MultiGit/'
    else:
        # See https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
        xdg_state_home: str
        if os.environ.get('XDG_STATE_HOME', ''):
            xdg_state_home = os.environ['XDG_STATE_HOME']
        else:
            xdg_state_home = os.path.expanduser('~/.local/state')

        multigit_log_dir = pathlib.Path(xdg_state_home) / 'MultiGit'

    # Make sure we are actually logging to a file which can be written
    try:
        if not multigit_log_dir.exists():
            multigit_log_dir.mkdir(parents=True)
        path_log_normal = multigit_log_dir / fname_log_normal
        path_log_debug = multigit_log_dir / fname_log_debug
        path_log_git_cmd = multigit_log_dir / fname_log_git_cmd

        if not is_writeable(path_log_normal) and is_writeable(path_log_debug):
            raise PermissionError('Can not create log file in location %s' % path_log_normal)
    except PermissionError:
        # is raised either by the mkdir() or the is_writeable() check
        # try yet another location
        temp_dir = pathlib.Path(tempfile.gettempdir())
        path_log_normal = temp_dir / fname_log_normal
        path_log_debug = temp_dir / fname_log_debug
        path_log_git_cmd = temp_dir / fname_log_git_cmd

        if not is_writeable(path_log_normal) and is_writeable(path_log_debug):
            # Desesperate case, do not create logs
            path_log_normal = None
            path_log_debug = None
            path_log_git_cmd = None

    mg_const.PATH_LOG_NORMAL = path_log_normal
    mg_const.PATH_LOG_DEBUG = path_log_debug
    mg_const.PATH_LOG_GIT_CMD = path_log_git_cmd


def init_logging(debug_activated: bool = False, run_from_tests: bool = False) -> None:
    '''run_from_tests: when activated, a different logging filename is used
    debug_activated: when True, a lot more informaiton is written to a debug log file.
                     The feature is activated when the tool is launched with --debug
        '''
    configure_logpath(debug_activated, run_from_tests)

    formatter = logging.Formatter('%(asctime)s %(name)15s:%(levelname)7s %(message)s')
    formatter.default_time_format = '%H:%M:%S'

    logger = logging.getLogger()

    # root log receives logging from all the tool
    # root log distributes log to:
    #   log file normal handler:
    #       filtered on level >= info
    #       exclude logger git command
    #   log file debug  handler: filtered on level debug
    #       filtered on level >= debug
    #       exclude logger git command
    #   stderr:
    #       filtered on level >= debug if in debug mode, >= info else
    #       exclude logger git command

    # configuration allows logging of debug messages
    if debug_activated:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


    if mg_const.PATH_LOG_NORMAL:
        fhandler_info = ConcurrentRotatingFileHandler(str(mg_const.PATH_LOG_NORMAL), encoding='utf8', maxBytes=10_000_000, backupCount=5)
        fhandler_info.setLevel( logging.INFO )
        fhandler_info.setFormatter(formatter)
        logger.addHandler(fhandler_info)
        # ensure that git_cmd records are not propagated to root handlers
        fhandler_info.addFilter(lambda record: record.name != mg_const.LOGGER_GIT_CMD)

    if (mg_const.PATH_LOG_DEBUG and debug_activated):
        fhandler_dbg = ConcurrentRotatingFileHandler(str(mg_const.PATH_LOG_DEBUG), encoding='utf8', maxBytes=10_000_000, backupCount=5)
        fhandler_dbg.setLevel( logging.DEBUG )
        fhandler_dbg.setFormatter(formatter)
        logger.addHandler(fhandler_dbg)
        # ensure that git_cmd records are not propagated to root handlers
        fhandler_dbg.addFilter(lambda record: record.name != mg_const.LOGGER_GIT_CMD)

    if mg_const.PATH_LOG_GIT_CMD:
        fhandler_git_cmd = ConcurrentRotatingFileHandler(str(mg_const.PATH_LOG_GIT_CMD), encoding='utf8', maxBytes=10_000_000, backupCount=5)
        fhandler_git_cmd.setLevel( logging.DEBUG )
        fhandler_git_cmd.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
        logger_git_cmd = logging.getLogger(mg_const.LOGGER_GIT_CMD)
        logger_git_cmd.addHandler(fhandler_git_cmd)


    shandler_err = logging.StreamHandler(sys.stderr)
    if debug_activated:
        shandler_err.setLevel( logging.DEBUG )
    else:
        shandler_err.setLevel( logging.INFO )
    logger.addHandler(shandler_err)
    shandler_err.setFormatter(formatter)


def main() -> None:
    if '--version' in sys.argv:
        print('Multigit v%s' % mg_const.VERSION)
        print('Based on:')
        print('- python v%s' % platform.python_version())
        try:
            from PySide2.QtCore import qVersion
            print('- Qt for Python v' + qVersion())
        except ImportError:
            print('- Qt for Python version not available')
        sys.exit(0)

    # to avoid crashes when Python exceptions are raised inside Qt slots
    sys.excepthook = handle_exception
    # sys.argv.append('--debug')
    debug_activated = False
    if '--debug' in sys.argv:
        debug_activated = True
        del sys.argv[sys.argv.index('--debug')]
    init_logging(debug_activated=debug_activated, run_from_tests=False)
    logging.info( 'Starting multigit v%s on %s' % (mg_const.VERSION, datetime.datetime.now().isoformat(sep=' ', timespec='seconds') ))
    try:
        from PySide2.QtCore import qVersion
        qt_version = 'v%s' % qVersion()
    except ImportError:
        qt_version = 'unknown'
    logging.info( 'Using Python v%s and Qt for Python %s' % (platform.python_version(), qt_version))
    main_gui()
    logging.info( 'Exit.' )


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_clone_from_mgit.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_button_history import MgButtonHistory

# import multigit_resources_rc

class Ui_CloneFromMgitFile(object):
    def setupUi(self, CloneFromMgitFile: QDialog) -> None:
        if not CloneFromMgitFile.objectName():
            CloneFromMgitFile.setObjectName(u"CloneFromMgitFile")
        CloneFromMgitFile.resize(935, 692)
        self.gridLayout_3 = QGridLayout(CloneFromMgitFile)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.groupBox = QGroupBox(CloneFromMgitFile)
        self.groupBox.setObjectName(u"groupBox")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        font.setWeight(75)
        self.groupBox.setFont(font)
        self.groupBox.setFlat(False)
        self.groupBox.setCheckable(False)
        self.gridLayout = QGridLayout(self.groupBox)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label = QLabel(self.groupBox)
        self.label.setObjectName(u"label")
        font1 = QFont()
        font1.setBold(False)
        font1.setWeight(50)
        self.label.setFont(font1)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 4)

        self.pushButtonChooseMgitFile = QPushButton(self.groupBox)
        self.pushButtonChooseMgitFile.setObjectName(u"pushButtonChooseMgitFile")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.pushButtonChooseMgitFile.sizePolicy().hasHeightForWidth())
        self.pushButtonChooseMgitFile.setSizePolicy(sizePolicy1)
        self.pushButtonChooseMgitFile.setMinimumSize(QSize(24, 0))
        self.pushButtonChooseMgitFile.setFont(font1)
        icon = QIcon()
        icon.addFile(u":/img/icons8-open-folder-64.png", QSize(), QIcon.Normal, QIcon.Off)
        self.pushButtonChooseMgitFile.setIcon(icon)

        self.gridLayout.addWidget(self.pushButtonChooseMgitFile, 1, 3, 1, 1)

        self.lineEditMgitFile = QLineEdit(self.groupBox)
        self.lineEditMgitFile.setObjectName(u"lineEditMgitFile")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        sizePolicy2.setHorizontalStretch(10)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.lineEditMgitFile.sizePolicy().hasHeightForWidth())
        self.lineEditMgitFile.setSizePolicy(sizePolicy2)
        self.lineEditMgitFile.setFont(font1)

        self.gridLayout.addWidget(self.lineEditMgitFile, 1, 1, 1, 1)

        self.historyButtonMgitFile = MgButtonHistory(self.groupBox)
        self.historyButtonMgitFile.setObjectName(u"historyButtonMgitFile")
        icon1 = QIcon()
        icon1.addFile(u":/img/icons8-history-64.png", QSize(), QIcon.Normal, QIcon.Off)
        self.historyButtonMgitFile.setIcon(icon1)

        self.gridLayout.addWidget(self.historyButtonMgitFile, 1, 2, 1, 1)


        self.gridLayout_3.addWidget(self.groupBox, 0, 0, 1, 2)

        self.groupBox_2 = QGroupBox(CloneFromMgitFile)
        self.groupBox_2.setObjectName(u"groupBox_2")
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setFont(font)
        self.groupBox_2.setFlat(False)
        self.groupBox_2.setCheckable(False)
        self.gridLayout_2 = QGridLayout(self.groupBox_2)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.pushButtonChooseDestDir = QPushButton(self.groupBox_2)
        self.pushButtonChooseDestDir.setObjectName(u"pushButtonChooseDestDir")
        sizePolicy1.setHeightForWidth(self.pushButtonChooseDestDir.sizePolicy().hasHeightForWidth())
        self.pushButtonChooseDestDir.setSizePolicy(sizePolicy1)
        self.pushButtonChooseDestDir.setMinimumSize(QSize(24, 0))
        self.pushButtonChooseDestDir.setFont(font1)
        self.pushButtonChooseDestDir.setIcon(icon)

        self.gridLayout_2.addWidget(self.pushButtonChooseDestDir, 1, 3, 1, 1)

        self.label_2 = QLabel(self.groupBox_2)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setFont(font1)

        self.gridLayout_2.addWidget(self.label_2, 0, 0, 1, 4)

        self.lineEditDestDir = QLineEdit(self.groupBox_2)
        self.lineEditDestDir.setObjectName(u"lineEditDestDir")
        sizePolicy2.setHeightForWidth(self.lineEditDestDir.sizePolicy().hasHeightForWidth())
        self.lineEditDestDir.setSizePolicy(sizePolicy2)
        self.lineEditDestDir.setFont(font1)

        self.gridLayout_2.addWidget(self.lineEditDestDir, 1, 1, 1, 1)

        self.historyButtonDestDir = MgButtonHistory(self.groupBox_2)
        self.historyButtonDestDir.setObjectName(u"historyButtonDestDir")
        self.historyButtonDestDir.setIcon(icon1)

        self.gridLayout_2.addWidget(self.historyButtonDestDir, 1, 2, 1, 1)


        self.gridLayout_3.addWidget(self.groupBox_2, 1, 0, 1, 2)

        self.groupBox_4 = QGroupBox(CloneFromMgitFile)
        self.groupBox_4.setObjectName(u"groupBox_4")
        self.groupBox_4.setFont(font)
        self.verticalLayout = QVBoxLayout(self.groupBox_4)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.radioDoNotAlterUrl = QRadioButton(self.groupBox_4)
        self.radioDoNotAlterUrl.setObjectName(u"radioDoNotAlterUrl")
        self.radioDoNotAlterUrl.setFont(font1)
        self.radioDoNotAlterUrl.setChecked(True)

        self.verticalLayout.addWidget(self.radioDoNotAlterUrl)

        self.radioStripUsername = QRadioButton(self.groupBox_4)
        self.radioStripUsername.setObjectName(u"radioStripUsername")
        self.radioStripUsername.setFont(font1)

        self.verticalLayout.addWidget(self.radioStripUsername)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.radioForceUsername = QRadioButton(self.groupBox_4)
        self.radioForceUsername.setObjectName(u"radioForceUsername")
        self.radioForceUsername.setFont(font1)

        self.horizontalLayout_2.addWidget(self.radioForceUsername)

        self.lineEditUsername = QLineEdit(self.groupBox_4)
        self.lineEditUsername.setObjectName(u"lineEditUsername")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(10)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.lineEditUsername.sizePolicy().hasHeightForWidth())
        self.lineEditUsername.setSizePolicy(sizePolicy3)
        self.lineEditUsername.setFont(font1)

        self.horizontalLayout_2.addWidget(self.lineEditUsername)

        self.historyButtonUsername = MgButtonHistory(self.groupBox_4)
        self.historyButtonUsername.setObjectName(u"historyButtonUsername")
        sizePolicy4 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.historyButtonUsername.sizePolicy().hasHeightForWidth())
        self.historyButtonUsername.setSizePolicy(sizePolicy4)
        self.historyButtonUsername.setIcon(icon1)

        self.horizontalLayout_2.addWidget(self.historyButtonUsername)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.gridLayout_3.addWidget(self.groupBox_4, 2, 0, 1, 1)

        self.groupBox_5 = QGroupBox(CloneFromMgitFile)
        self.groupBox_5.setObjectName(u"groupBox_5")
        self.groupBox_5.setFont(font)
        self.verticalLayout_5 = QVBoxLayout(self.groupBox_5)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.label_4 = QLabel(self.groupBox_5)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setFont(font1)

        self.verticalLayout_5.addWidget(self.label_4)

        self.radioButtonDirExistsGitFail = QRadioButton(self.groupBox_5)
        self.radioButtonDirExistsGitFail.setObjectName(u"radioButtonDirExistsGitFail")
        self.radioButtonDirExistsGitFail.setFont(font1)
        self.radioButtonDirExistsGitFail.setChecked(True)

        self.verticalLayout_5.addWidget(self.radioButtonDirExistsGitFail)

        self.radioButtonDirExistsSkipDir = QRadioButton(self.groupBox_5)
        self.radioButtonDirExistsSkipDir.setObjectName(u"radioButtonDirExistsSkipDir")
        self.radioButtonDirExistsSkipDir.setFont(font1)

        self.verticalLayout_5.addWidget(self.radioButtonDirExistsSkipDir)

        self.radioButtonDirExistsDelDir = QRadioButton(self.groupBox_5)
        self.radioButtonDirExistsDelDir.setObjectName(u"radioButtonDirExistsDelDir")
        self.radioButtonDirExistsDelDir.setFont(font1)

        self.verticalLayout_5.addWidget(self.radioButtonDirExistsDelDir)


        self.gridLayout_3.addWidget(self.groupBox_5, 2, 1, 1, 1)

        self.tabWidgetProject = QTabWidget(CloneFromMgitFile)
        self.tabWidgetProject.setObjectName(u"tabWidgetProject")
        self.tabProject = QWidget()
        self.tabProject.setObjectName(u"tabProject")
        self.verticalLayout_3 = QVBoxLayout(self.tabProject)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.labelProjDesc = QLabel(self.tabProject)
        self.labelProjDesc.setObjectName(u"labelProjDesc")
        sizePolicy5 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy5.setHorizontalStretch(0)
        sizePolicy5.setVerticalStretch(1)
        sizePolicy5.setHeightForWidth(self.labelProjDesc.sizePolicy().hasHeightForWidth())
        self.labelProjDesc.setSizePolicy(sizePolicy5)
        self.labelProjDesc.setFont(font)

        self.horizontalLayout_3.addWidget(self.labelProjDesc)

        self.textEditProjectDesc = QTextEdit(self.tabProject)
        self.textEditProjectDesc.setObjectName(u"textEditProjectDesc")
        sizePolicy6 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy6.setHorizontalStretch(0)
        sizePolicy6.setVerticalStretch(1)
        sizePolicy6.setHeightForWidth(self.textEditProjectDesc.sizePolicy().hasHeightForWidth())
        self.textEditProjectDesc.setSizePolicy(sizePolicy6)
        self.textEditProjectDesc.setMaximumSize(QSize(16777215, 50))
        self.textEditProjectDesc.setReadOnly(True)

        self.horizontalLayout_3.addWidget(self.textEditProjectDesc)


        self.verticalLayout_3.addLayout(self.horizontalLayout_3)

        self.treeWidgetRepoList = QTreeWidget(self.tabProject)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)
        self.treeWidgetRepoList.setColumnCount(4)

        self.verticalLayout_3.addWidget(self.treeWidgetRepoList)

        self.tabWidgetProject.addTab(self.tabProject, "")
        self.tabConfFile = QWidget()
        self.tabConfFile.setObjectName(u"tabConfFile")
        self.verticalLayout_2 = QVBoxLayout(self.tabConfFile)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.textEditProj = QTextEdit(self.tabConfFile)
        self.textEditProj.setObjectName(u"textEditProj")

        self.verticalLayout_2.addWidget(self.textEditProj)

        self.tabWidgetProject.addTab(self.tabConfFile, "")

        self.gridLayout_3.addWidget(self.tabWidgetProject, 3, 0, 1, 2)

        self.groupBox_3 = QGroupBox(CloneFromMgitFile)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.horizontalLayout = QHBoxLayout(self.groupBox_3)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.buttonBox = QDialogButtonBox(self.groupBox_3)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel)

        self.horizontalLayout.addWidget(self.buttonBox)


        self.gridLayout_3.addWidget(self.groupBox_3, 4, 0, 1, 2)

        QWidget.setTabOrder(self.lineEditMgitFile, self.historyButtonMgitFile)
        QWidget.setTabOrder(self.historyButtonMgitFile, self.pushButtonChooseMgitFile)
        QWidget.setTabOrder(self.pushButtonChooseMgitFile, self.lineEditDestDir)
        QWidget.setTabOrder(self.lineEditDestDir, self.historyButtonDestDir)
        QWidget.setTabOrder(self.historyButtonDestDir, self.pushButtonChooseDestDir)
        QWidget.setTabOrder(self.pushButtonChooseDestDir, self.lineEditUsername)
        QWidget.setTabOrder(self.lineEditUsername, self.historyButtonUsername)
        QWidget.setTabOrder(self.historyButtonUsername, self.radioButtonDirExistsGitFail)
        QWidget.setTabOrder(self.radioButtonDirExistsGitFail, self.radioButtonDirExistsSkipDir)
        QWidget.setTabOrder(self.radioButtonDirExistsSkipDir, self.radioButtonDirExistsDelDir)
        QWidget.setTabOrder(self.radioButtonDirExistsDelDir, self.tabWidgetProject)
        QWidget.setTabOrder(self.tabWidgetProject, self.textEditProjectDesc)
        QWidget.setTabOrder(self.textEditProjectDesc, self.treeWidgetRepoList)
        QWidget.setTabOrder(self.treeWidgetRepoList, self.textEditProj)

        self.retranslateUi(CloneFromMgitFile)
        self.buttonBox.accepted.connect(CloneFromMgitFile.accept)
        self.buttonBox.rejected.connect(CloneFromMgitFile.reject)
        self.radioForceUsername.toggled.connect(self.lineEditUsername.setEnabled)
        self.radioForceUsername.toggled.connect(self.historyButtonUsername.setEnabled)

        self.tabWidgetProject.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(CloneFromMgitFile)
    # setupUi

    def retranslateUi(self, CloneFromMgitFile: QDialog) -> None:
        CloneFromMgitFile.setWindowTitle(QCoreApplication.translate("CloneFromMgitFile", u"Clone from Multigit file", None))
        self.groupBox.setTitle(QCoreApplication.translate("CloneFromMgitFile", u"Multigit File", None))
        self.label.setText(QCoreApplication.translate("CloneFromMgitFile", u"Select a multigit file (extension .mgit)", None))
#if QT_CONFIG(tooltip)
        self.pushButtonChooseMgitFile.setToolTip(QCoreApplication.translate("CloneFromMgitFile", u"Choose multigit file", None))
#endif // QT_CONFIG(tooltip)
        self.pushButtonChooseMgitFile.setText("")
#if QT_CONFIG(tooltip)
        self.historyButtonMgitFile.setToolTip(QCoreApplication.translate("CloneFromMgitFile", u"Select recent multigit file", None))
#endif // QT_CONFIG(tooltip)
        self.historyButtonMgitFile.setText("")
        self.groupBox_2.setTitle(QCoreApplication.translate("CloneFromMgitFile", u"Project Destination Directory", None))
#if QT_CONFIG(tooltip)
        self.pushButtonChooseDestDir.setToolTip(QCoreApplication.translate("CloneFromMgitFile", u"Choose directory", None))
#endif // QT_CONFIG(tooltip)
        self.pushButtonChooseDestDir.setText("")
        self.label_2.setText(QCoreApplication.translate("CloneFromMgitFile", u"Choose target directory for the clone:", None))
#if QT_CONFIG(tooltip)
        self.historyButtonDestDir.setToolTip(QCoreApplication.translate("CloneFromMgitFile", u"Select recent directory", None))
#endif // QT_CONFIG(tooltip)
        self.historyButtonDestDir.setText("")
        self.groupBox_4.setTitle(QCoreApplication.translate("CloneFromMgitFile", u"User Name in url", None))
        self.radioDoNotAlterUrl.setText(QCoreApplication.translate("CloneFromMgitFile", u"Do not alter url", None))
        self.radioStripUsername.setText(QCoreApplication.translate("CloneFromMgitFile", u"Remove login from url if present", None))
        self.radioForceUsername.setText(QCoreApplication.translate("CloneFromMgitFile", u"Force login username to :", None))
#if QT_CONFIG(tooltip)
        self.historyButtonUsername.setToolTip(QCoreApplication.translate("CloneFromMgitFile", u"Select recent user name", None))
#endif // QT_CONFIG(tooltip)
        self.historyButtonUsername.setText("")
        self.groupBox_5.setTitle(QCoreApplication.translate("CloneFromMgitFile", u"Existing directory behavior", None))
        self.label_4.setText(QCoreApplication.translate("CloneFromMgitFile", u"If clone directory already exists :", None))
        self.radioButtonDirExistsGitFail.setText(QCoreApplication.translate("CloneFromMgitFile", u"Git fails to clone", None))
        self.radioButtonDirExistsSkipDir.setText(QCoreApplication.translate("CloneFromMgitFile", u"Skip directory", None))
        self.radioButtonDirExistsDelDir.setText(QCoreApplication.translate("CloneFromMgitFile", u"Delete directory", None))
        self.labelProjDesc.setText(QCoreApplication.translate("CloneFromMgitFile", u"Project description :", None))
        self.textEditProjectDesc.setHtml(QCoreApplication.translate("CloneFromMgitFile", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:'Noto Sans'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'MS Shell Dlg 2'; font-size:8.25pt;\"><br /></p></body></html>", None))
        ___qtreewidgetitem = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("CloneFromMgitFile", u"URL", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("CloneFromMgitFile", u"Path", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("CloneFromMgitFile", u"HEAD target", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("CloneFromMgitFile", u"Repository", None));
        self.tabWidgetProject.setTabText(self.tabWidgetProject.indexOf(self.tabProject), QCoreApplication.translate("CloneFromMgitFile", u"Repositories", None))
        self.tabWidgetProject.setTabText(self.tabWidgetProject.indexOf(self.tabConfFile), QCoreApplication.translate("CloneFromMgitFile", u"Multigit file content", None))
        self.groupBox_3.setTitle("")
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_multigit_widget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree
from src.mg_button_history import MgButtonHistory

# import multigit_resources_rc

class Ui_MultigitWidget(object):
    def setupUi(self, MultigitWidget: QWidget) -> None:
        if not MultigitWidget.objectName():
            MultigitWidget.setObjectName(u"MultigitWidget")
        MultigitWidget.resize(649, 639)
        self.verticalLayout_2 = QVBoxLayout(MultigitWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.splitter = QSplitter(MultigitWidget)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Vertical)
        self.layoutWidget = QWidget(self.splitter)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.verticalLayout = QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(self.layoutWidget)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.lineEditBaseDir = QLineEdit(self.layoutWidget)
        self.lineEditBaseDir.setObjectName(u"lineEditBaseDir")
        self.lineEditBaseDir.setReadOnly(True)

        self.horizontalLayout.addWidget(self.lineEditBaseDir)

        self.buttonHistoryBaseDir = MgButtonHistory(self.layoutWidget)
        self.buttonHistoryBaseDir.setObjectName(u"buttonHistoryBaseDir")
        icon = QIcon()
        icon.addFile(u":/img/icons8-history-64.png", QSize(), QIcon.Normal, QIcon.Off)
        self.buttonHistoryBaseDir.setIcon(icon)

        self.horizontalLayout.addWidget(self.buttonHistoryBaseDir)

        self.buttonOpenBaseDir = QPushButton(self.layoutWidget)
        self.buttonOpenBaseDir.setObjectName(u"buttonOpenBaseDir")
        icon1 = QIcon()
        icon1.addFile(u":/img/icons8-open-folder-64.png", QSize(), QIcon.Normal, QIcon.Off)
        self.buttonOpenBaseDir.setIcon(icon1)

        self.horizontalLayout.addWidget(self.buttonOpenBaseDir)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.repoTree = MgRepoTree(self.layoutWidget)
        self.repoTree.headerItem().setText(0, "")
        self.repoTree.setObjectName(u"repoTree")
        self.repoTree.setLineWidth(3)
        self.repoTree.setMidLineWidth(4)
        self.repoTree.setAlternatingRowColors(True)
        self.repoTree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.repoTree.setRootIsDecorated(False)
        self.repoTree.setUniformRowHeights(True)
        self.repoTree.setItemsExpandable(False)
        self.repoTree.setSortingEnabled(True)
        self.repoTree.setAllColumnsShowFocus(True)

        self.verticalLayout.addWidget(self.repoTree)

        self.splitter.addWidget(self.layoutWidget)
        self.tabWidget = QTabWidget(self.splitter)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tabLastCommit = QWidget()
        self.tabLastCommit.setObjectName(u"tabLastCommit")
        self.horizontalLayout_3 = QHBoxLayout(self.tabLastCommit)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.textEditCommit = QTextEdit(self.tabLastCommit)
        self.textEditCommit.setObjectName(u"textEditCommit")
        self.textEditCommit.setReadOnly(True)

        self.horizontalLayout_3.addWidget(self.textEditCommit)

        self.tabWidget.addTab(self.tabLastCommit, "")
        self.tabModifiedFiles = QWidget()
        self.tabModifiedFiles.setObjectName(u"tabModifiedFiles")
        self.horizontalLayout_4 = QHBoxLayout(self.tabModifiedFiles)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.textEditModFiles = QTextEdit(self.tabModifiedFiles)
        self.textEditModFiles.setObjectName(u"textEditModFiles")
        font = QFont()
        font.setFamily(u"Courier New")
        self.textEditModFiles.setFont(font)
        self.textEditModFiles.setReadOnly(True)

        self.horizontalLayout_4.addWidget(self.textEditModFiles)

        self.tabWidget.addTab(self.tabModifiedFiles, "")
        self.splitter.addWidget(self.tabWidget)

        self.verticalLayout_2.addWidget(self.splitter)

        QWidget.setTabOrder(self.lineEditBaseDir, self.buttonHistoryBaseDir)
        QWidget.setTabOrder(self.buttonHistoryBaseDir, self.buttonOpenBaseDir)
        QWidget.setTabOrder(self.buttonOpenBaseDir, self.repoTree)
        QWidget.setTabOrder(self.repoTree, self.tabWidget)
        QWidget.setTabOrder(self.tabWidget, self.textEditCommit)
        QWidget.setTabOrder(self.textEditCommit, self.textEditModFiles)

        self.retranslateUi(MultigitWidget)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MultigitWidget)
    # setupUi

    def retranslateUi(self, MultigitWidget: QWidget) -> None:
        MultigitWidget.setWindowTitle(QCoreApplication.translate("MultigitWidget", u"Form", None))
        self.label.setText(QCoreApplication.translate("MultigitWidget", u"Base directory :", None))
#if QT_CONFIG(tooltip)
        self.buttonHistoryBaseDir.setToolTip(QCoreApplication.translate("MultigitWidget", u"Open recent base directory", None))
#endif // QT_CONFIG(tooltip)
        self.buttonHistoryBaseDir.setText("")
#if QT_CONFIG(tooltip)
        self.buttonOpenBaseDir.setToolTip(QCoreApplication.translate("MultigitWidget", u"Open base directory", None))
#endif // QT_CONFIG(tooltip)
        self.buttonOpenBaseDir.setText("")
        ___qtreewidgetitem = self.repoTree.headerItem()
        ___qtreewidgetitem.setText(4, QCoreApplication.translate("MultigitWidget", u"Last Remote Synchro", None));
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("MultigitWidget", u"Status", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("MultigitWidget", u"Head", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("MultigitWidget", u"Git Repo Path", None));
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabLastCommit), QCoreApplication.translate("MultigitWidget", u"Last Commit", None))
        self.textEditModFiles.setHtml(QCoreApplication.translate("MultigitWidget", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:'Courier New'; font-size:9.75pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:'MS Shell Dlg 2'; font-size:8pt; color:#0000ff;\">@@ this is a description @@</span><span style=\" font-family:'MS Shell Dlg 2'; font-size:8pt;\"> bla bla bla<br />this line is plain<br /></span><span style=\" font-family:'MS Shell Dlg 2'; font-size:8pt; color:#aa0000;\">- this line is removed</span><span style=\" font-family:'MS Shell Dlg 2'; font-size:8pt;\"><br /></span><span style=\" font-family:'MS Shell Dlg 2'; font-size:8pt; color:#00aa00;\">+ this line is added</span><span style=\" font-family:'MS Shell Dlg 2'; font-size:8"
                        "pt;\"><br /><br /></span></p></body></html>", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabModifiedFiles), QCoreApplication.translate("MultigitWidget", u"Modified Files", None))
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import Any, TYPE_CHECKING, Optional, cast, Tuple
import logging

from PySide2.QtGui import QIcon, QColor, QPalette
from PySide2.QtWidgets import QTreeWidgetItem, QApplication, QWidget, QTreeWidget

from src.mg_utils import extractInt
from src import mg_config as mgc
from src.mg_const import COL_UPDATE, COL_REPO_NAME, COL_HEAD, COL_STATUS, COL_REMOTE_SYNCHRO, DISPLAY_IN_BOLD_MSG, \
    DISPLAY_IN_ITALIC_MSG, COL_SHA1, SHORT_SHA1_NB_DIGITS, COL_URL, MSG_LOCAL_BRANCH, MSG_LOCAL_BRANCH_TOOLTIP
from src.mg_exec_task_item import getIcon, IconSet
from src.mg_repo_info import MgRepoInfo

dbg = logging.getLogger('mg_repo_tree_item').debug
error = logging.getLogger('mg_repo_tree_item').error

DELTA_ROTATION = 30
TOTAL_ROTATION = 360

from src.mg_utils import istrcmp

def shallBoldItalizeSynchro(remote_synchro: str) -> Tuple[bool, bool]:
    '''Return whether message shall be in bold and in italic'''
    doBoldSynchro = False
    for msg in DISPLAY_IN_BOLD_MSG:
        if msg in remote_synchro:
            doBoldSynchro = True
            break

    doItalicSynchro = remote_synchro in DISPLAY_IN_ITALIC_MSG
    return (doBoldSynchro, doItalicSynchro)


class MgRepoTreeItem(QTreeWidgetItem):

    def __init__(self, repoInfo: MgRepoInfo, *args: Any) -> None:
        super().__init__(*args)
        if self.text(COL_REPO_NAME) == '':
            # we are completely empty, fill with minimalistic information
            self.setText(COL_UPDATE, '')
            self.setText(COL_REPO_NAME, repoInfo.name)
            self.setText(COL_HEAD, '...')
            self.setText(COL_STATUS, '...')
            self.setText(COL_REMOTE_SYNCHRO, '...')

        f = self.font(COL_SHA1)
        f.setFixedPitch(True)
        self.setFont(COL_SHA1, f)

        self.repoInfo = repoInfo
        self.repoInfo.repo_info_available.connect(self.slotRepoInfoAvailable)
        self.repoInfo.repo_update_in_progress.connect(self.slotRepoUpdateInProgress)
        self.repoInfo.repo_deleted.connect(self.slotRepoDeleted)


    def markItemInProgress(self) -> None:
        try:
            dbg('markItemInProgress(%s)' % self.repoInfo.name)
            self.setText(COL_UPDATE, '')
            self.setIcon(COL_UPDATE, QIcon(':img/icons8-loader-96.png'))
        except RuntimeError:
            # happens when updating an item, which is a Python object, but the C++
            # underlying object has already been deleted
            # just ignore the error
            error('Trying to update an item which has already been deleted!')
            error('Just ignore the error for now but we should find a better way!')
            return
        QApplication.processEvents()


    def slotRepoDeleted(self, repo_name: str) -> None:
        '''The repo mapped to this item was deleted'''
        dbg('MgRepoTreeItem.slotRepoDeleted(%s)' % repo_name)
        treeWidget = self.treeWidget()
        # note: when item has already been removed from MgTreeWidget, this is None

        if treeWidget is not None:
            idx = treeWidget.indexOfTopLevelItem(self)
            if idx == -1:
                dbg('Item already deleted')
                return
            treeWidget.takeTopLevelItem(idx)


    def slotRepoUpdateInProgress(self, repo_name: str) -> None:
        '''Called when the update of the repo info has completed'''
        dbg('slotRepoUpdateInProgress(%s)' % repo_name)
        if repo_name != self.repoInfo.name:
            error('Item slotRepoUpdateInProgress() of repo %s is called with argument %s' % (self.repoInfo.name, repo_name))
            error('This happens usually when the base directory is changed while repositories are still fetching information')
            error('We should try a better way to avoid this in the future!')
            return
        self.markItemInProgress()


    def slotRepoInfoAvailable(self, repo_name: str) -> None:
        '''Called when the update of the repo info has completed'''
        dbg('slotRepoInfoAvailable(%s)' % repo_name)

        # sometimes, the view is modified, the item is deleted but we still
        # receive the signal from Python.
        try:
            # just to trigger access to C++ object
            self.setExpanded(self.isExpanded())
        except RuntimeError:
            error('Trying to update an item which has already been deleted!')
            error('Just ignore the error for now but we should find a better way!')
            return

        if repo_name != self.repoInfo.name:
            raise AssertionError('Item slotRepoInfoAvailable() of repo %s is called with argument %s' % (self.repoInfo.name, repo_name))
        self.fillRepoItem()


    def fillRepoItem(self) -> None:
        '''Fill a QTreeWidgetItem from the associated repoInfo'''
        dbg('fillRepoItem(%s, ...)' % self.repoInfo.name)
        self.setIcon(COL_UPDATE, getIcon(IconSet.Empty))
        self.setText(COL_UPDATE, '')
        self.setText(COL_REPO_NAME, self.repoInfo.name)
        self.setText(COL_HEAD, self.repoInfo.head)
        self.setText(COL_STATUS, self.repoInfo.status)
        self.setText(COL_REMOTE_SYNCHRO, self.repoInfo.remote_synchro )
        if self.repoInfo.remote_synchro == MSG_LOCAL_BRANCH:
            # add a tooltip to give an easy-fix for the missing remote branch
            self.setToolTip(COL_REMOTE_SYNCHRO, MSG_LOCAL_BRANCH_TOOLTIP)
        else:
            self.setToolTip(COL_REMOTE_SYNCHRO, '')
        self.setText(COL_SHA1, (self.repoInfo.commit_sha1 or '')[:SHORT_SHA1_NB_DIGITS])
        self.setText(COL_URL, self.repoInfo.url or '')

        doBoldStatus = self.repoInfo.status != 'OK'
        f = self.font(COL_STATUS)
        f.setBold(doBoldStatus)
        self.setFont(COL_STATUS, f)

        doBoldSynchro, doItalicSynchro = shallBoldItalizeSynchro(self.repoInfo.remote_synchro)
        f = self.font(COL_REMOTE_SYNCHRO)
        f.setBold(doBoldSynchro)
        f.setItalic(doItalicSynchro)
        self.setFont(COL_REMOTE_SYNCHRO, f)

        doColorHeadBranch = 'branch' in self.repoInfo.head
        doColorHeadTag    = 'tag'    in self.repoInfo.head
        if doColorHeadTag or doColorHeadBranch:
            colorRgb = self.foreground(COL_HEAD).color()
            if doColorHeadBranch:
                colorRgb = mgc.get_config_instance().get(mgc.CONFIG_HEAD_COLOR_BRANCH, mgc.DEFAULT_CONFIG_HEAD_COLOR_BRANCH)
            elif doColorHeadTag:
                colorRgb = mgc.get_config_instance().get(mgc.CONFIG_HEAD_COLOR_TAG, mgc.DEFAULT_CONFIG_HEAD_COLOR_TAG)
            self.setForeground(COL_HEAD, QColor(colorRgb))
        else:
            self.setForeground(COL_HEAD, QApplication.palette().brush(QPalette.Active, QPalette.WindowText ))

        MgRepoTreeItem.autoAdjustColumnSize(self.treeWidget())

        try:
            # we refresh the columns one by one, to give priority to the other repo still processing their basic information
            # So, we don't request URL before SHA1 is available
            if not self.treeWidget().isColumnHidden(COL_SHA1):
                self.repoInfo.ensure_sha1(self.cbSha1available)

            elif not self.treeWidget().isColumnHidden(COL_URL):
                self.repoInfo.ensure_url(self.cbUrlAvailable)
        except RuntimeError:
            error('Trying to update an item which has already been deleted!')
            error('Just ignore the error for now but we should find a better way!')
            return



    def cbSha1available(self, commit_sha1: str) -> None:
        '''Called when sha1 information is eventually available'''

        # sometimes, the view is modified, the item is deleted but we still
        # receive the signal from Python.
        try:
            self.setText(COL_SHA1, commit_sha1[:SHORT_SHA1_NB_DIGITS])
            self.autoAdjustColumnSize(self.treeWidget())

            if not self.treeWidget().isColumnHidden(COL_URL):
                self.repoInfo.ensure_url(self.cbUrlAvailable)

        except RuntimeError:
            error('Trying to update an item which has already been deleted!')
            error('Just ignore the error for now but we should find a better way!')
            return



    def cbUrlAvailable(self, url: str) -> None:
        '''Called when url information is eventually available'''

        # sometimes, the view is modified, the item is deleted but we still
        # receive the signal from Python.
        try:
            # just to trigger access to C++ object
            self.setExpanded(self.isExpanded())
        except RuntimeError:
            error('Trying to update an item which has already been deleted!')
            error('Just ignore the error for now but we should find a better way!')
            return

        self.setText(COL_URL, url)
        self.autoAdjustColumnSize(self.treeWidget())


    @staticmethod
    def autoAdjustColumnSize(treeWidget: QTreeWidget) -> None:
        '''Adjust automatically the column size to the largest item'''
        for i in range(treeWidget.columnCount()):
            treeWidget.resizeColumnToContents(i)
        QApplication.processEvents()

    def __lt__(self, other: 'QTreeWidgetItem') -> bool:
        col = self.treeWidget().sortColumn()
        if not col in [COL_STATUS, COL_REMOTE_SYNCHRO]:
            # regular sorting
            return istrcmp(self.text(col), other.text(col))

        colTextSelf = self.text(col)
        colTextOther = other.text(col)
        if len(colTextSelf) and len(colTextOther) and colTextSelf[0].isdigit() and colTextOther[0].isdigit():
            # natural number sorting if we can
            return extractInt(colTextSelf) < extractInt(colTextOther)

        # regular sort strategy will compare strings and place all number starting strings before others
        return istrcmp(self.text(col), other.text(col))



